# -*- coding: utf-8 -*-

import nibabel as nib
import numpy as np
from dipy.core.gradients import gradient_table
from dipy.segment.mask import median_otsu
from dipy.denoise.nlmeans import nlmeans
import dipy.denoise.noise_estimate as ne
import dipy.reconst.dti as dti
from dipy.reconst.dti import mean_diffusivity, color_fa, lower_triangular
from dipy.data import get_sphere
# from dipy.viz import fvtk
from time import time
from os.path import expanduser, join
from scipy.ndimage.morphology import binary_dilation

from dipy.reconst.shm import CsaOdfModel
from dipy.direction import peaks_from_model

from dipy.tracking.local import ThresholdTissueClassifier
from dipy.tracking import utils
from dipy.tracking.local import LocalTracking
from dipy.io.trackvis import save_trk


def load_volume(path,name):
	print('Loading volume ...')
	fvol = join(path,name)
	vol = nib.load(fvol)
	data = vol.get_data()
	print('Data dimensions: ')
	print(data.shape)
	print('Voxel size: ')
	print(vol.get_header().get_zooms()[:3])
	return vol

def load_g_table(path):
	print('Gradients table ...')
	fbvals = join(path,'bvals')
	fbvecs = join(path,'bvecs')
	gtab = gradient_table(fbvals,fbvecs)
	print('Gradient Table')
	print(gtab.info)
	return gtab

def crop_volume(vol, gtab):
	print('Cropping the volume to the necessary data ...')
	# identify b0s and first shell
	bindx = np.where( gtab.bvals < 1500) 

	# crop them into a new volume
	bvals1k0 = gtab.bvals[ bindx ]
	bvecs1k0 = np.squeeze(gtab.bvecs[ bindx, : ])
	gtab1k0 = gradient_table(bvals1k0,bvecs1k0)
	print(gtab1k0.info)

	np.savetxt('bvals',bvals1k0)
	np.savetxt('bvecs',bvecs1k0)

	# create new volume and bvals and bvecs files
	data = vol.get_data()
	crp_data = np.squeeze(data[:,:,:,bindx])
	print('Data dimensions: ')
	print(crp_data.shape)

	crp_vol = nib.Nifti1Image(crp_data.astype(np.float32), vol.get_affine())
	nib.save(crp_vol, 'data_crop.nii.gz')

	return crp_vol, gtab1k0

def brain_extraction(vol):
	print('Extracting the brain ...')
	###### PLAY WITH THRESHOLDING PARAMETERS ######
	# b0_mask, mask = median_otsu(data,2,1)
	# b0_mask, mask = median_otsu(data,4,4,autocrop=True)
	vol_masked, mask = median_otsu(vol.get_data(),4,1,autocrop=True)

	mask_img = nib.Nifti1Image(mask.astype(np.float32), vol.get_affine())
	vol_masked_img = nib.Nifti1Image(vol_masked.astype(np.float32), vol.get_affine())

	fname = 'bmcdmri16'
	nib.save(mask_img, fname + '_binary_mask.nii.gz')
	nib.save(vol_masked_img, fname + '_mask.nii.gz')
	return vol_masked_img, mask_img

def denoise_vol(vol, mask):
	print('Denoising the data. It will take a while ...')
	t = time()
	# SENSE Reconstruction N = 1
	sigma = ne.estimate_sigma(vol.get_data(), N=1)
	den = nlmeans(data, sigma=sigma, mask=mask.get_data())
	print("total time", time() - t)
	print("vol size", den.shape)

	den_vol = nib.Nifti1Image(den, vol.get_affine())
	nib.save(den_vol, 'denoised.nii.gz')
	return den_vol

def dit_fit(vol, gtab):
	print('Fitting the Tensor Model ...')
	tenmodel = dti.TensorModel(gtab)
	tenfit = tenmodel.fit(vol.get_data())
	return tenfit

def restore_fit(vol, gtab):
	print('Fitting the Tensor Model with RESTORE ...')
	sigma = ne.estimate_sigma(vol.get_data(),N=1)
	tenmodel = dti.TensorModel(gtab, fit_method='RESTORE',sigma=sigma)
	tenfit = tenmodel.fit(vol.get_data())
	return tenfit

def calc_dti_metrics(vol, tenfit, name):
	print('Saving anisotropy measures (FA, MD, RGB) into files ...')
	FA = tenfit.fa
	FA[np.isnan(FA)] = 0
	fa_vol = nib.Nifti1Image(FA.astype(np.float32), vol.get_affine())
	nib.save(fa_vol, name + '_tensor_fa.nii.gz')

	evecs_vol = nib.Nifti1Image(tenfit.evecs.astype(np.float32), vol.get_affine())
	nib.save(evecs_vol, name + '_tensor_evecs.nii.gz')

	MD = dti.mean_diffusivity(tenfit.evals)	
	nib.save(nib.Nifti1Image(MD.astype(np.float32), vol.get_affine()), name + '_tensors_md.nii.gz')

	return


def tractography(vol,gtab,mask):
	csa_model = CsaOdfModel(gtab, sh_order=4)
	sphere = get_sphere('symmetric724')


	csa_peaks = peaks_from_model(model=csa_model,
                                     data=vol.get_data(),
                                     sphere=sphere,
                                     relative_peak_threshold=.5,
                                     min_separation_angle=25,
                                     mask=mask.get_data(),
                                     return_odf=False,
                                     normalize_peaks=True,
                                     npeaks=5,
                                     parallel=True,
                                     nbr_processes=2)

	classifier = ThresholdTissueClassifier(csa_peaks.gfa, .25)

	seed_mask = labels == 2
	seeds = utils.seeds_from_mask(seed_mask, density=[2, 2, 2], affine=vol.get_affine())

	# Initialization of LocalTracking. The computation happens in the next step.
	streamlines = LocalTracking(csa_peaks, classifier, seeds, vol.get_affine(), step_size=.5)
	# Compute streamlines and store as a list.
	streamlines = list(streamlines)
	save_trk("CSA_detr.trk", streamlines, vol.get_affine(), labels.shape)
	return


def main():

	## PREPROCESSING ##

	home = expanduser('~')
	droot = join(home,'Workspace','Diffusion Exercise BMC', 'data')
	wspath = join(droot,'HCP_100307','T1w','Diffusion')

	# load volume and gradients
	vol = load_volume(wspath,'data.nii.gz')
	gtab = load_g_table(wspath)

	# Crop the volume and gradients to keep the 1000 s/mm^2 shell
	# and the reverse polarity acquisition
	crp_vol, crp_g_table = crop_volume(vol, gtab)

	# brain extraction
	bex_vol, bex_mask = brain_extraction(crp_vol)

	# denoising
	dn_vol = denoise_vol(bex_vol, bex_mask)

	 ## FIRST EXCERCISE: DTI METRICS ##

	# The slice to visualize must contain the CC for reference
	slice = 67
	# tensor on original images
	orig_DTI_tenfit = dit_fit(bex_vol, crp_g_table)
	calc_dti_metrics(bex_vol, orig_DTI_tenfit, 'original_dti')
	# visualize_ellipsoids(orig_DTI_tenfit, slice,'original_dti')

	# tensor on denoised images
	dn_DTI_tenfit = dit_fit(dn_vol, crp_g_table)
	calc_dti_metrics(dn_vol, dn_DTI_tenfit, 'denoised_dti')
	# visualize_ellipsoids(dn_DTI_tenfit, slice,'denoised_dti')

	# RESTORE on original images
	orig_RST_tenfit = restore_fit(bex_vol, crp_g_table)
	calc_dti_metrics(dn_vol, dn_DTI_tenfit, 'original_restore')
	# visualize_ellipsoids(orig_RST_tenfit, slice,'original_restore')

	# RESTORE on denoised images
	dn_RST_tenfit = restore_fit(dn_vol, crp_g_table)
	calc_dti_metrics(dn_vol, dn_RST_tenfit, 'denoised_restore')
	# visualize_ellipsoids(dn_RST_tenfit, slice,'denoised_restore')

	## SECOND EXCERCISE: TRACTOGRAPHY BASED ON SPHERICAL HARMONICS ##

	# tractography on denoised data 
	tractography(dn_vol,crp_g_table,bex_mask)




if __name__ == "__main__":
	main()
